# -*- coding: utf-8 -*-

# Copyright 2019 FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Dataclass utility functions for Spock"""

from copy import copy
from typing import List
from typing import Tuple
from spock.backend.dataclass._dataclasses import field


def cast(x):
    """Recasts lists as tuples

    *Args*:

        x: object

    *Returns*:

        x: object or object recast as Tuple
    """
    if isinstance(x, list):
        x = tuple(x)
    return x


def _def_list(values: List):
    """Creates a list of default values for List datatype that is mutable

    *Args*:

        values: default list

    Returns:

        list built from default factory

    """
    return field(default_factory=lambda: copy(values))


def _def_tuple(values: Tuple):
    """Creates a tuple of default values for Tuple datatype that is mutable

        *Args*:

            values: default tuple

        Returns:

            tuple built from default factory

        """
    return field(default_factory=lambda: copy(values))



# -*- coding: utf-8 -*-

# Copyright 2019 FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles the definitions of arguments types for Spock (backend: attrs)"""

import sys
from enum import EnumMeta
from typing import TypeVar
from typing import Union
import attr
minor = sys.version_info.minor
if minor < 7:
    from typing import GenericMeta as _GenericAlias
else:
    from typing import _GenericAlias


class SavePath(str):
    """Spock special key for saving the Spock config to file

    Defines a special key use to save the current Spock config to file

    """
    def __new__(cls, x):
        return super().__new__(cls, x)


def _extract_base_type(typed):
    """Extracts the name of the type from a _GenericAlias

    *Args*:

        typed: the type of the parameter

    *Returns*:

        name of type
    """
    if hasattr(typed, '__args__'):
        if minor < 7:
            name = typed.__name__
        else:
            name = typed._name
        bracket_val = f"{name}[{_extract_base_type(typed.__args__[0])}]"
        return bracket_val
    else:
        bracket_value = typed.__name__
    return bracket_value


def _recursive_generic_validator(typed):
    """Recursively assembles the validators for nested generic types

    Walks through the nested type structure and determines whether to recurse all the way to a base type. Once it
    hits the base type it bubbles up the correct validator that is nested within the upper validator

    *Args*:

        typed: input type

    *Returns*:

        return_type: recursively built deep_iterable validators

    """
    if hasattr(typed, '__args__'):
        # If there are more __args__ then we still need to recurse as it is still a GenericAlias
        return_type = attr.validators.deep_iterable(
            member_validator=_recursive_generic_validator(typed.__args__[0]),
            iterable_validator=attr.validators.instance_of(typed.__origin__)
        )
        return return_type
    else:
        # If no more __args__ then we are to the base type and need to bubble up the type
        return_type = attr.validators.instance_of(typed)
    return return_type


def _generic_alias_katra(typed, default=None, optional=False):
    """Private interface to create a subscripted generic_alias katra

    A 'katra' is the basic functional unit of `spock`. It defines a parameter using attrs as the backend, type checks
    both simple types and subscripted GenericAlias types (e.g. lists and tuples), handles setting default parameters,
    and deals with parameter optionality

    Handles: List[type] and Tuple[type]

    *Args*:

        typed: the type of the parameter to define
        default: the default value to assign if given
        optional: whether to make the parameter optional or not (thus allowing None)

    *Returns*:

        x: Attribute from attrs

    """
    # base python class from which a GenericAlias is derived
    base_typed = typed.__origin__
    if default is not None:
        x = attr.ib(validator=_recursive_generic_validator(typed), default=default, type=base_typed,
                    metadata={'base': _extract_base_type(typed), 'type': typed})
        # x = attr.ib(validator=_recursive_generic_iterator(typed), default=default, type=base_typed,
        #             metadata={'base': _extract_base_type(typed)})
    elif optional:
        # if there's no default, but marked as optional, then set the default to None
        x = attr.ib(validator=attr.validators.optional(_recursive_generic_validator(typed)), type=base_typed,
                    default=default, metadata={'optional': True, 'base': _extract_base_type(typed), 'type': typed})
        # x = attr.ib(validator=attr.validators.optional(_recursive_generic_iterator(typed)), type=base_typed,
        #             default=default, metadata={'optional': True, 'base': _extract_base_type(typed)})
    else:
        x = attr.ib(validator=_recursive_generic_validator(typed), type=base_typed,
                    metadata={'base': _extract_base_type(typed), 'type': typed})
        # x = attr.ib(validator=_recursive_generic_iterator(typed), type=base_typed,
        #             metadata={'base': _extract_base_type(typed)})
    return x


def _enum_katra(typed, default=None, optional=False):
    """Private interface to create a Enum typed katra

    A 'katra' is the basic functional unit of `spock`. It defines a parameter using attrs as the backend, type checks
    both simple types and subscripted GenericAlias types (e.g. lists and tuples), handles setting default parameters,
    and deals with parameter optionality

    *Args*:

        typed: the type of the parameter to define
        default: the default value to assign if given
        optional: whether to make the parameter optional or not (thus allowing None)

    *Returns*:

        x: Attribute from attrs

    """
    # First check if the types of Enum are the same
    type_set = {type(val.value) for val in typed}
    if len(type_set) > 1:
        raise TypeError(f"Enum cannot be defined with multiple types: {type_set}")
    base_type = list(type_set)[-1]
    allowed = [val.value for val in typed]
    if default is not None:
        x = attr.ib(
            validator=[attr.validators.instance_of(base_type), attr.validators.in_(allowed)],
            default=default, type=base_type, metadata={'base': typed.__name__})
    elif optional:
        x = attr.ib(
            validator=attr.validators.optional([attr.validators.instance_of(base_type), attr.validators.in_(allowed)]),
            default=default, type=base_type, metadata={'base': typed.__name__})
    else:
        x = attr.ib(validator=[attr.validators.instance_of(base_type), attr.validators.in_(allowed)], type=typed,
                    metadata={'base': typed.__name__})
    return x


def _type_katra(typed, default=None, optional=False):
    """Private interface to create a simple typed katra

    A 'katra' is the basic functional unit of `spock`. It defines a parameter using attrs as the backend, type checks
    both simple types and subscripted GenericAlias types (e.g. lists and tuples), handles setting default parameters,
    and deals with parameter optionality

    Handles: bool, string, float, int, List, and Tuple

    *Args*:

        typed: the type of the parameter to define
        default: the default value to assign if given
        optional: whether to make the parameter optional or not (thus allowing None)

    *Returns*:

        x: Attribute from attrs

    """
    # Grab the name first based on if it is a base type or GenericAlias
    if isinstance(typed, type):
        name = typed.__name__
    elif isinstance(typed, _GenericAlias):
        if minor < 7:
            name = typed.__name__
        else:
            name = typed._name
    else:
        raise TypeError('Encountered an uxpected type in _type_katra')
    special_key = None
    # Default booleans to false and optional due to the nature of a boolean
    if isinstance(typed, type) and name == "bool":
        optional = True
        if default is not True:
            default = False
    # For the save path type we need to swap the type back to it's base class (str)
    elif isinstance(typed, type) and name == "SavePath":
        optional = True
        special_key = name
        typed = str
    if default is not None:
        # if a default is provided, that takes precedence
        x = attr.ib(validator=attr.validators.instance_of(typed), default=default, type=typed,
                    metadata={'base': name, 'special_key': special_key})
    elif optional:
        x = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(typed)), default=default, type=typed,
                    metadata={'optional': True, 'base': name, 'special_key': special_key})
    else:
        x = attr.ib(validator=attr.validators.instance_of(typed), type=typed, metadata={'base': name,
                                                                                        'special_key': special_key})
    return x


def _handle_optional_typing(typed):
    """Handles when a type hint is Optional

    Handles Optional[type] typing and strips out the base type to pass back to the creation of a katra which needs base
    typing

    *Args*:

        typed: type

    *Returns*:

        typed: type (modified if Optional)
        optional: boolean for katra creation
    """
    # Set optional to false
    optional = False
    # Check if it has __args__ to look for optionality as it is a GenericAlias
    if hasattr(typed, '__args__'):
        # If it is more than one than it is most likely optional but check against NoneType in the tuple to verify
        # Check the length of type __args__
        type_args = typed.__args__
        # Optional[X] has type_args = (X, None) and is equal to Union[X, None]
        if (len(type_args) == 2) and (typed == Union[type_args[0], None]):
            # Since this is true we need to strip out the OG type
            # Grab all the types that are not NoneType and collapse to a list
            type_list = [val for val in type_args if val is not type(None)]
            if len(type_list) > 1:
                raise TypeError(f"Passing multiple subscript types to GenericAlias is not supported: {type_list}")
            else:
                typed = type_list[0]
            # Set the optional flag to true
            optional = True
    return typed, optional


def katra(typed, default=None):
    """Public interface to create a katra

    A 'katra' is the basic functional unit of `spock`. It defines a parameter using attrs as the backend, type checks
    both simple types and subscripted GenericAlias types (e.g. lists and tuples), handles setting default parameters,
    and deals with parameter optionality

    *Args*:

        typed: the type of the parameter to define
        default: the default value to assign if given
        optional: whether to make the parameter optional or not (thus allowing None)

    Returns:

        x: Attribute from attrs

    """
    # Handle optionals
    typed, optional = _handle_optional_typing(typed)
    # We need to check if the type is a _GenericAlias so that we can handle subscripted general types
    # If it is subscript typed it will not be T which python uses as a generic type name
    if isinstance(typed, _GenericAlias) and (not isinstance(typed.__args__[0], TypeVar)):
        x = _generic_alias_katra(typed=typed, default=default, optional=optional)
    elif isinstance(typed, EnumMeta):
        x = _enum_katra(typed=typed, default=default, optional=optional)
    else:
        x = _type_katra(typed=typed, default=default, optional=optional)
    return x



# -*- coding: utf-8 -*-

# Copyright 2019 FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Utility functions for Spock"""

import ast
from enum import EnumMeta
import os
import subprocess
import sys
from time import localtime
from time import strftime
import git
minor = sys.version_info.minor
if minor < 7:
    from typing import GenericMeta as _GenericAlias
else:
    from typing import _GenericAlias


def make_argument(arg_name, arg_type, parser):
    """Make argparser argument based on type

    Based on the type passed in handle the creation of the argparser argument so that overrides will have the correct
    behavior when set

    *Args*:

        arg_name: name for the argument
        arg_type: type of the argument
        parser: current parser

    Returns:

        parser: updated argparser

    """
    # For generic alias we take the input string and use a custom type callable to convert
    if isinstance(arg_type, _GenericAlias):
        parser.add_argument(arg_name, required=False, type=_handle_generic_type_args)
    # For choice enums we need to check a few things first
    elif isinstance(arg_type, EnumMeta):
        type_set = list({type(val.value) for val in arg_type})[0]
        parser.add_argument(arg_name, required=False, type=type_set)
    # For booleans we map to store true
    elif arg_type == bool:
        parser.add_argument(arg_name, required=False, action='store_true')
    # Else we are a simple base type which we can cast to
    else:
        parser.add_argument(arg_name, required=False, type=arg_type)
    return parser


def _handle_generic_type_args(val):
    return ast.literal_eval(val)


def add_info(out_dict):
    """Adds extra information to the output dictionary

    *Args*:

        out_dict: output dictionary

    *Returns*:

        out_dict: output dictionary
    """
    out_dict = add_date_info(out_dict)
    out_dict = add_repo_info(out_dict)
    return out_dict


def make_blank_git(out_dict):
    """Adds blank git info

    *Args*:

        out_dict: current output dictionary

    Returns:

        out_dict: output dictionary with added git info

    """
    for key in ('BRANCH', 'COMMIT SHA', 'STATUS', 'ORIGIN'):
        out_dict.update({f'# Git {key}': 'UNKNOWN'})
    return out_dict


def add_repo_info(out_dict):
    """Adds GIT information to the output dictionary

    *Args*:

        out_dict: output dictionary

    *Returns*:

        out_dict: output dictionary
    """
    try:
        # Assume we are working out of a repo
        repo = git.Repo(os.getcwd())
        # Check if we are really in a detached head state as this will fail
        if minor < 7:
            head_result = subprocess.run('git rev-parse --abbrev-ref --symbolic-full-name HEAD', stdout=subprocess.PIPE,
                                         shell=True)
        else:
            head_result = subprocess.run('git rev-parse --abbrev-ref --symbolic-full-name HEAD', capture_output=True,
                                         shell=True)
        if head_result.stdout.decode().rstrip('\n') == 'HEAD':
            out_dict = make_blank_git(out_dict)
        else:
            out_dict.update({'# Git BRANCH': repo.active_branch.name})
            out_dict.update({'# Git COMMIT SHA': repo.head.object.hexsha})
            if len(repo.untracked_files) > 0 or len(repo.head.commit.diff(None)) > 0:
                git_status = 'DIRTY'
            else:
                git_status = 'CLEAN'
            out_dict.update({'# Git STATUS': git_status})
            out_dict.update({'# Git ORIGIN': repo.remotes.origin.url})
    except git.InvalidGitRepositoryError:
        # But it's okay if we are not
        out_dict = make_blank_git(out_dict)

    return out_dict


def add_date_info(out_dict):
    """Adds date information to the output dictionary

    *Args*:

        out_dict: output dictionary

    *Returns*:

        out_dict: output dictionary
    """
    out_dict.update({'# Run Date': strftime('%Y_%m_%d_%H_%M_%S', localtime())})
    return out_dict



# -*- coding: utf-8 -*-

# Copyright 2019 FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Creates the spock config decorator that wraps dataclasses"""

from spock.backend.dataclass.config import spock_dataclass
from spock.backend.attr.config import spock_attr

# Dataclasses for legacy support -- now wraps attr via an adapter
spock_config = spock_dataclass

# Simplified decorator for attrs
spock = spock_attr



# -*- coding: utf-8 -*-

# Copyright 2019 FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles prepping and saving the Spock config"""

import attr
from spock.backend.base import BaseSaver


class AttrSaver(BaseSaver):
    """Base class for saving configs for the attrs backend

    Contains methods to build a correct output payload and then writes to file based on the file
    extension

    *Attributes*:

        _writers: maps file extension to the correct i/o handler

    """
    def __init__(self):
        super().__init__()

    def __call__(self, *args, **kwargs):
        return AttrSaver()

    def _clean_up_values(self, payload, extra_info, file_extension):
        out_dict = {}
        for key, val in vars(payload).items():
            # Append comment tag to the base class and convert the spock class to a dict
            if file_extension == '.json':
                out_dict.update({key: attr.asdict(val)})
            else:
                out_dict.update({('# ' + key): attr.asdict(val)})
        # Convert values
        clean_dict = self._clean_output(out_dict, extra_info)
        return clean_dict



# -*- coding: utf-8 -*-

# Copyright 2019 FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles the building/saving of the configurations from the Spock config classes"""

from pathlib import Path
import attr
from spock.backend.attr.builder import AttrBuilder
from spock.backend.attr.payload import AttrPayload
from spock.backend.attr.saver import AttrSaver
# from spock.backend.dataclass.builder import DataClassBuilder
# from spock.backend.dataclass.payload import DataClassPayload
# from spock.backend.dataclass.saver import DataClassSaver
# from spock.backend.dataclass._dataclasses import is_dataclass


class ConfigArgBuilder:
    """Automatically generates dataclass instances from config file(s)

    This class builds out necessary arguments from *args classes, reads
    the arguments from specified config file(s), and subsequently (via chained
    call to generate) generates each class instance based on the necessary
    field values for each backend class instance

    *Attributes*:

        _arg_namespace: generated argument namespace
        _builder_obj: instance of a BaseBuilder class
        _create_save_path: boolean to make the path to save to
        _dict_args: dictionary args from the command line
        _payload_obj: instance of a BasePayload class
        _saver_obj: instance of a BaseSaver class

    """
    def __init__(self, *args, configs=None, create_save_path=False, desc='', no_cmd_line=False, **kwargs):
        backend = self._set_backend(args)
        self._create_save_path = create_save_path
        self._builder_obj = backend.get('builder')(
            *args, configs=configs, create_save_path=create_save_path, desc=desc, no_cmd_line=no_cmd_line, **kwargs)
        self._payload_obj = backend.get('payload')
        self._saver_obj = backend.get('saver')
        try:
            self._dict_args = self._get_payload()
            self._arg_namespace = self._builder_obj.generate(self._dict_args)
        except Exception as e:
            self._builder_obj.print_usage_and_exit(str(e), sys_exit=False)
            raise ValueError(e)

    def __call__(self, *args, **kwargs):
        """Call to self to allow chaining

        *Args*:

            *args:
            **kwargs:

        *Returns*:

            ConfigArgBuilder: self instance
        """
        return ConfigArgBuilder(*args, **kwargs)

    def generate(self):
        """Generate method that returns the actual argument namespace

        *Returns*:

            argument namespace consisting of all config classes

        """
        return self._arg_namespace

    @staticmethod
    def _set_backend(args):
        """Determines which backend class to use

        *Args*:

            args: list of classes passed to the builder

        *Returns*:

            backend: class of backend

        """
        # Gather if all attr backend
        type_attrs = all([attr.has(arg) for arg in args])
        if not type_attrs:
            raise TypeError("*args must be of all attrs backend")
        elif type_attrs:
            backend = {'builder': AttrBuilder, 'payload': AttrPayload, 'saver': AttrSaver}
        else:
            raise TypeError("*args must be of all attrs backend")
        return backend

    def _get_config_paths(self):
        """Get config paths from all methods

        Config paths can enter from either the command line or be added in the class init call
        as a kwarg (configs=[])

        *Returns*:

            args: namespace of args

        """
        # Call the objects get_config_paths function
        args = self._builder_obj.get_config_paths()
        return args

    def _get_payload(self):
        """Get the parameter payload from the config file(s)

        Calls the various ways to get configs and then parses to retrieve the parameter payload

        *Returns*:

            payload: dictionary of parameter values

        """
        args = self._get_config_paths()
        if args.help:
            self._builder_obj.print_usage_and_exit()
        payload = {}
        for configs in args.config:
            payload.update(self._payload_obj().payload(self._builder_obj.input_classes, configs, args))
        return payload

    def save(self, user_specified_path=None, extra_info=True, file_extension='.yaml'):
        """Saves the current config setup to file with a UUID

        *Args*:

            user_specified_path: if user provides a path it will be used as the path to write
            extra_info: additional info to write to saved config (run date and git info)
            file_extension: file type to write (default: yaml)

        *Returns*:

            self so that functions can be chained
        """
        if user_specified_path is not None:
            save_path = Path(user_specified_path)
        elif self._builder_obj.save_path is not None:
            save_path = Path(self._builder_obj.save_path)
        else:
            raise ValueError('Save did not receive a valid path from: (1) markup file(s) or (2) '
                             'the keyword arg user_specified_path')
        # Call the saver class and save function
        self._saver_obj().save(self._arg_namespace, save_path, self._create_save_path, extra_info, file_extension)
        return self



# -*- coding: utf-8 -*-

# Copyright 2019 FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles the building/saving of the configurations from the Spock config classes"""

import sys
import attr
from spock.backend.base import BaseBuilder


class AttrBuilder(BaseBuilder):
    """Attr specific builder

    Class that handles building for the attr backend

    *Attributes*

        input_classes: list of input classes that link to a backend
        _configs: None or List of configs to read from
        _create_save_path: boolean to make the path to save to
        _desc: description for the arg parser
        _no_cmd_line: flag to force no command line reads
        save_path: list of path(s) to save the configs to

    """
    def __init__(self, *args, configs=None, create_save_path=False, desc='', no_cmd_line=False, **kwargs):
        super().__init__(*args, configs=configs, create_save_path=create_save_path, desc=desc,
                         no_cmd_line=no_cmd_line, **kwargs)
        for arg in self.input_classes:
            if not attr.has(arg):
                raise TypeError('*arg inputs to ConfigArgBuilder must all be class instances with attrs attributes')

    def print_usage_and_exit(self, msg=None, sys_exit=True):
        print('USAGE:')
        print(f'  {sys.argv[0]} -c [--config] config1 [config2, config3, ...]')
        print('CONFIG:')
        for attrs_class in self.input_classes:
            print('  ' + attrs_class.__name__ + ':')
            for val in attrs_class.__attrs_attrs__:
                type_string = val.metadata['base']
                # Construct the type with the metadata
                if 'optional' in val.metadata:
                    type_string = "Optional[{0}]".format(type_string)
                print(f'    {val.name}: {type_string}')
        if msg is not None:
            print(msg)
        if sys_exit:
            sys.exit(1)

    def _handle_arguments(self, args, class_obj):
        attr_name = class_obj.__name__
        fields = {}
        for val in class_obj.__attrs_attrs__:
            # Check if namespace is named and then check for key -- checking for local class def
            if attr_name in args and val.name in args[attr_name]:
                fields[val.name] = args[attr_name][val.name]
            # If not named then just check for keys -- checking for global def
            elif val.name in args:
                fields[val.name] = args[val.name]
            # Check for special keys to set
            if 'special_key' in val.metadata and val.metadata['special_key'] is not None:
                if val.name in args:
                    self.save_path = args[val.name]
                elif val.default is not None:
                    self.save_path = val.default
        return fields



# -*- coding: utf-8 -*-

# Copyright 2019 FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""I/O handlers for various file formats"""

from abc import ABC
from abc import abstractmethod
import json
import re
import toml
import yaml


class Handler(ABC):
    """Base class for file type loaders

    ABC for loaders

    """
    @abstractmethod
    def load(self, path):
        """Load function for file type

        *Args*:

            path: path to file

        *Returns*:


        """
        raise NotImplementedError

    @abstractmethod
    def save(self, out_dict, path):
        """Write function for file type

        *Args*:

            out_dict: payload to write
            path: path to write out

        *Returns*:

        """
        raise NotImplementedError


class YAMLHandler(Handler):
    """YAML class for loading YAML config files

    Base YAML class

    """
    # override default SafeLoader behavior to correctly
    # interpret 1e1 (as opposed to 1.e+1) as 10
    # https://stackoverflow.com/questions/30458977/yaml-loads-5e-6-as-string-and-not-a-number/30462009#30462009
    yaml.SafeLoader.add_implicit_resolver(
        u'tag:yaml.org,2002:float',
        re.compile(u'''^(?:
         [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
        |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
        |\\.[0-9_]+(?:[eE][-+][0-9]+)?
        |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
        |[-+]?\\.(?:inf|Inf|INF)
        |\\.(?:nan|NaN|NAN))$''', re.X),
        list(u'-+0123456789.')
    )

    def load(self, path):
        """YAML load function

        *Args*:

            path: path to YAML file

        *Returns*:

            base_payload: dictionary of read file

        """
        file_contents = open(path, 'r').read()
        file_contents = re.sub(r'--([a-zA-Z0-9_]*)', r'\g<1>: True', file_contents)
        base_payload = yaml.safe_load(file_contents)
        return base_payload

    def save(self, out_dict, path):
        """Write function for YAML type

        *Args*:

            out_dict: payload to write
            path: path to write out

        *Returns*:

        """
        yaml.dump(out_dict, path, default_flow_style=False)


class TOMLHandler(Handler):
    """TOML class for loading TOML config files

    Base TOML class

    """
    def load(self, path):
        """TOML load function

        *Args*:

            path: path to TOML file

        Returns:

            base_payload: dictionary of read file

        """
        base_payload = toml.load(path)
        return base_payload

    def save(self, out_dict, path):
        """Write function for TOML type

        *Args*:

            out_dict: payload to write
            path: path to write out

        *Returns*:

        """
        toml.dump(out_dict, path)


class JSONHandler(Handler):
    """JSON class for loading JSON config files

    Base JSON class

    """
    def load(self, path):
        """JSON load function

        *Args*:

            path: path to JSON file

        Returns:

            base_payload: dictionary of read file

        """
        with open(path) as json_fid:
            base_payload = json.load(json_fid)
        return base_payload

    def save(self, out_dict, path):
        """Write function for JSON type

        *Args*:

            out_dict: payload to write
            path: path to write out

        *Returns*:

        """
        with open(path.name, 'w') as json_fid:
            json.dump(out_dict, json_fid, indent=4, separators=(',', ': '))



# -*- coding: utf-8 -*-

# Copyright 2019 FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles payloads from markup files"""

from itertools import chain
from spock.backend.base import BasePayload


class DataClassPayload(BasePayload):
    """Handles building the payload for dataclass backend

    This class builds out the payload from config files of multiple types. It handles various
    file types and also composition of config files via a recursive calls

    *Attributes*:

        _loaders: maps of each file extension to the loader class

    """
    def __init__(self):
        super().__init__()

    def __call__(self, *args, **kwargs):
        """Call to allow self chaining

        *Args*:

            *args:
            **kwargs:

        *Returns*:

            Payload: instance of self

        """
        return DataClassPayload()

    @staticmethod
    def _update_payload(base_payload, input_classes, payload):
        # Get basic args
        dc_fields = {dc.__name__: list(vars(dc).get('__dataclass_fields__').keys()) for dc in input_classes}
        # Get the choice args and insert them
        # dc_fields = self._handle_choices(dc_fields, input_classes)
        for keys, values in base_payload.items():
            # check if the keys, value pair is expected by a dataclass
            if keys != 'config':
                # Dict infers that we are overriding a global setting in a specific config
                if isinstance(values, dict):
                    # we're in a namespace
                    # Check for incorrect specific override of global def
                    if keys not in dc_fields:
                        raise TypeError(f'Referring to a class space {keys} that is undefined')
                    for i_keys in values.keys():
                        if i_keys not in dc_fields[keys]:
                            raise ValueError(f'Provided an unknown argument named {keys}.{i_keys}')
                else:
                    # Chain all the values from multiple spock classes into one list
                    if keys not in list(chain(*dc_fields.values())):
                        raise ValueError(f'Provided an unknown argument named {keys}')
            if keys in payload and isinstance(values, dict):
                payload[keys].update(values)
            else:
                payload[keys] = values
        return payload



# -*- coding: utf-8 -*-

# Copyright 2019 FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles base Spock classes"""

from abc import ABC
from abc import abstractmethod
import argparse
import os
from pathlib import Path
import sys
from uuid import uuid1
from spock.handlers import JSONHandler
from spock.handlers import TOMLHandler
from spock.handlers import YAMLHandler
from spock.utils import add_info
from spock.utils import make_argument


class BaseSaver(ABC):  # pylint: disable=too-few-public-methods
    """Base class for saving configs

    Contains methods to build a correct output payload and then writes to file based on the file
    extension

    *Attributes*:

        _writers: maps file extension to the correct i/o handler

    """
    def __init__(self):
        self._writers = {'.yaml': YAMLHandler, '.toml': TOMLHandler, '.json': JSONHandler}

    def save(self, payload, path, create_save_path=False, extra_info=True, file_extension='.yaml'):  #pylint: disable=too-many-arguments
        """Writes Spock config to file

        Cleans and builds an output payload and then correctly writes it to file based on the
        specified file extension

        *Args*:

            payload: current config payload
            path: path to save
            create_save_path: boolean to create the path if non-existent
            extra_info: boolean to write extra info
            file_extension: what type of file to write

        *Returns*:

            None

        """
        supported_extensions = list(self._writers.keys())
        if file_extension not in list(self._writers.keys()):
            raise ValueError(f'Invalid fileout extension. Expected a fileout from {supported_extensions}')
        # Make the filename
        name = str(uuid1()) + '.spock.cfg' + file_extension
        fid = path / name
        # Fix up values
        out_dict = self._clean_up_values(payload, extra_info, file_extension)
        try:
            if not os.path.exists(path) and create_save_path:
                os.makedirs(path)
            with open(fid, 'w') as file_out:
                self._writers.get(file_extension)().save(out_dict, file_out)
        except OSError as e:
            print(f'Not a valid file path to write to: {fid}')
            raise e

    @abstractmethod
    def _clean_up_values(self, payload, extra_info, file_extension):
        """Clean up the config payload so it can be written to file

        *Args*:

            payload: dirty payload
            extra_info: boolean to add extra info
            file_extension: type of file to write

        *Returns*:

            clean_dict: cleaned output payload

        """

    @staticmethod
    def _clean_output(out_dict, extra_info):
        """Clean up the dictionary so it can be written to file

        *Args*:

            out_dict: cleaned dictionary
            extra_info: boolean to add extra info

        *Returns*:

            clean_dict: cleaned output payload

        """
        # Convert values
        clean_dict = {}
        for key, val in out_dict.items():
            clean_inner_dict = {}
            for inner_key, inner_val in val.items():
                # Convert tuples to lists so they get written correctly
                if isinstance(inner_val, tuple):
                    clean_inner_dict.update({inner_key: list(inner_val)})
                elif inner_val is not None:
                    clean_inner_dict.update({inner_key: inner_val})
            clean_dict.update({key: clean_inner_dict})
        if extra_info:
            clean_dict = add_info(clean_dict)
        return clean_dict


class BaseBuilder(ABC):  # pylint: disable=too-few-public-methods
    """Base class for building the backend specific builders

    This class handles the interface to the backend with the generic ConfigArgBuilder so that different
    backends can be used to handle processing

    *Attributes*

        input_classes: list of input classes that link to a backend
        _configs: None or List of configs to read from
        _create_save_path: boolean to make the path to save to
        _desc: description for the arg parser
        _no_cmd_line: flag to force no command line reads
        save_path: list of path(s) to save the configs to

    """
    def __init__(self, *args, configs=None, create_save_path=False, desc='', no_cmd_line=False, **kwargs):
        self.input_classes = args
        self._configs = configs
        self._create_save_path = create_save_path
        self._desc = desc
        self._no_cmd_line = no_cmd_line
        self.save_path = None

    @abstractmethod
    def print_usage_and_exit(self, msg=None, sys_exit=True):
        """Prints the help message and exits

        *Args*:

            msg: message to print pre exit

        *Returns*:

            None

        """

    @abstractmethod
    def _handle_arguments(self, args, class_obj):
        """Handles all argument mapping

        Creates a dictionary of named parameters that are mapped to the final type of object

        *Args*:

            args: read file arguments
            class_obj: instance of a class obj

        *Returns*:

            fields: dictionary of mapped parameters

        """

    def generate(self, dict_args):
        """Method to auto-generate the actual class instances from the generated args

        Based on the generated arguments groups and the args read in from the config file(s)
        this function instantiates the classes with the necessary field or attr values

        *Args*:

            dict_args: dictionary of arguments from the configs

        *Returns*:

            namespace containing automatically generated instances of the classes
        """
        auto_dict = {}
        for attr_classes in self.input_classes:
            attr_build = self._auto_generate(dict_args, attr_classes)
            auto_dict.update({type(attr_build).__name__: attr_build})
        return argparse.Namespace(**auto_dict)

    def _auto_generate(self, args, input_class):
        """Builds an instance of a DataClass

        Builds an instance of a dataclass with the necessary field values from the argument
        dictionary read from the config file(s)

        *Args*:

            args: dictionary of arguments read from the config file(s)
            data_class: data class to build

        *Returns*:

            An instance of data_class with correct values assigned to fields
        """
        # Handle the basic data types
        fields = self._handle_arguments(args, input_class)
        return input_class(**fields)

    def get_config_paths(self):
        """Get config paths from all methods

        Config paths can enter from either the command line or be added in the class init call
        as a kwarg (configs=[])

        *Returns*:

            args: namespace of args

        """
        # Check if the no_cmd_line is not flagged and if the configs are not empty

        if self._no_cmd_line and (self._configs is None):
            raise ValueError("Flag set for preventing command line read but no paths were passed to the config kwarg")
        if not self._no_cmd_line:
            args = self._build_override_parsers(desc=self._desc)
        else:
            args = argparse.Namespace(config=[], help=False)
        if self._configs is not None:
            args = self._get_from_kwargs(args, self._configs)
        return args

    def _build_override_parsers(self, desc):
        """Creates parsers for command-line overrides

        Builds the basic command line parser for configs and help, iterates through all defined attr to make
        a general override parser, and then iterates through each attr instance to make namespace specific override
        parsers

        *Args*:

            desc: argparser description

        *Returns*:

            args: argument namespace

        """
        parser = argparse.ArgumentParser(description=desc, add_help=False)
        parser.add_argument('-c', '--config', required=False, nargs='+', default=[])
        parser.add_argument('-h', '--help', action='store_true')
        # Build out a general parser for parent level attr
        parser = self._make_general_override_parser(parser=parser, input_classes=self.input_classes)
        # Build out each class specific parser
        for val in self.input_classes:
            parser = self._make_group_override_parser(parser=parser, class_obj=val)
        # args = parser.parse_args()
        args, _ = parser.parse_known_args(sys.argv)
        return args

    def _make_general_override_parser(self, parser, input_classes):
        """Makes a general level override parser

        Flattens all the attrs into a single dictionary and makes a general level parser for the attr name

        *Args*:

            parser: argument parser
            input_classes: list of input classes for a specific backend

        *Returns*:

            parser: argument parser with new general overrides

        """
        # Make all names list
        all_attr = {}
        for class_obj in input_classes:
            for val in class_obj.__attrs_attrs__:
                val_type = val.metadata['type'] if 'type' in val.metadata else val.type
                if hasattr(all_attr, val.name):
                    if all_attr[val.name] is not val_type:
                        print(f"Warning: Ignoring general override for {val.name} as the class specific types differ")
                else:
                    all_attr.update({val.name: val_type})
        self._check_protected_keys(all_attr)
        group_parser = parser.add_argument_group(title="General Overrides")
        for k, v in all_attr.items():
            arg_name = '--' + k
            group_parser = make_argument(arg_name, v, group_parser)
        return parser

    @staticmethod
    def _make_group_override_parser(parser, class_obj):
        """Makes a name specific override parser for a given class obj

        Takes a class object of the backend and adds a new argument group with argument names given with name
        Class.name so that individual parameters specific to a class can be overridden.

        *Args*:

            parser: argument parser
            class_obj: instance of a backend class

        *Returns*:

            parser: argument parser with new class specific overrides

        """
        attr_name = class_obj.__name__
        group_parser = parser.add_argument_group(title=str(attr_name) + " Specific Overrides")
        for val in class_obj.__attrs_attrs__:
            val_type = val.metadata['type'] if 'type' in val.metadata else val.type
            arg_name = '--' + str(attr_name) + '.' + val.name
            group_parser = make_argument(arg_name, val_type, group_parser)
        return parser

    @staticmethod
    def _check_protected_keys(all_attr):
        """Test for protected keys

        Tests to see if an attribute has been defined at the genreral level that is within the protected list that
        would break basic command line handling.

        Args:
            all_attr: dictionary of all attr

        """
        protected_names = ['config', 'help']
        if any([val in all_attr.keys() for val in protected_names]):
            raise ValueError(f"Using a protected name from {protected_names} at general class level which prevents "
                             f"command line overrides")

    @staticmethod
    def _get_from_arg_parser(desc):
        """Get configs from command line

        Gets the config file(s) from the command line arg parser

        *Args*:

            desc: description text for the cmd line argparser

        *Returns*:

            args: namespace of command line args

        """
        # Pull in args via the arg parser pointing to the config file
        parser = argparse.ArgumentParser(description=desc, add_help=False)
        parser.add_argument('-c', '--config', required=False, nargs='+', default=[])
        parser.add_argument('-h', '--help', action='store_true')
        args, _ = parser.parse_known_args(sys.argv)
        return args

    @staticmethod
    def _get_from_kwargs(args, configs):
        """Get configs from the configs kwarg


        *Args*:

            args: argument namespace
            configs: config kwarg

        *Returns*:

            args: arg namespace

        """
        if type(configs).__name__ == 'list':
            args.config.extend(configs)
        else:
            raise TypeError('configs kwarg must be of type list')
        return args


class BasePayload(ABC):  # pylint: disable=too-few-public-methods
    """Handles building the payload for config file(s)

    This class builds out the payload from config files of multiple types. It handles various
    file types and also composition of config files via a recursive calls

    *Attributes*:

        _loaders: maps of each file extension to the loader class

    """
    def __init__(self):
        self._loaders = {'.yaml': YAMLHandler(), '.toml': TOMLHandler(), '.json': JSONHandler()}

    @staticmethod
    @abstractmethod
    def _update_payload(base_payload, input_classes, payload):
        """Updates the payload

        Checks the parameters defined in the config files against the provided classes and if
        passable adds them to the payload

        *Args*:

            base_payload: current payload
            input_classes: class to roll into
            payload: total payload

        *Returns*:

            payload: updated payload

        """

    def payload(self, input_classes, path, cmd_args):
        """Builds the payload from config files

        Public exposed call to build the payload and set any command line overrides

        *Args*:

            input_classes: list of backend classes
            path: path to config file(s)
            cmd_args: command line overrides

        *Returns*:

            payload: dictionary of all mapped parameters

        """
        payload = self._payload(input_classes, path)
        payload = self._handle_overrides(payload, cmd_args)
        return payload

    def _payload(self, input_classes, path):
        """Private call to construct the payload

        Main function call that builds out the payload from config files of multiple types. It handles
        various file types and also composition of config files via a recursive calls

        *Args*:
            input_classes: list of backend classes
            path: path to config file(s)

        *Returns*:

            payload: dictionary of all mapped parameters

        """
        # Match to loader based on file-extension
        config_extension = Path(path).suffix.lower()
        supported_extensions = list(self._loaders.keys())
        if config_extension not in supported_extensions:
            raise TypeError(f'File extension {config_extension} not supported\n'
                            f'Must be from {supported_extensions}')
        # Load from file
        base_payload = self._loaders.get(config_extension).load(path)
        payload = {}
        if 'config' in base_payload:
            payload = self._handle_includes(
                base_payload, config_extension, input_classes, path, payload)
        payload = self._update_payload(base_payload, input_classes, payload)
        return payload

    def _handle_includes(self, base_payload, config_extension, input_classes, path, payload):  # pylint: disable=too-many-arguments
        """Handles config composition

        For all of the config tags in the config file this function will recursively call the payload function
        with the composition path to get the additional payload(s) from the composed file(s)

        *Args*:

            base_payload: base payload that has a config kwarg
            config_extension: file type
            input_classes: defined backend classes
            path: path to base file
            payload: payload pulled from composed files

        *Returns*:

            payload: payload update from composed files

        """
        included_params = {}
        for inc_path in base_payload['config']:
            if not os.path.exists(inc_path):
                # maybe it's relative?
                abs_inc_path = os.path.join(os.path.dirname(path), inc_path)
            else:
                abs_inc_path = inc_path
            if not os.path.exists(abs_inc_path):
                raise RuntimeError(f'Could not find included {config_extension} file {inc_path}!')
            included_params.update(self._payload(input_classes, abs_inc_path))
        payload.update(included_params)
        return payload

    def _handle_overrides(self, payload, args):
        """Handle command line overrides

        Iterate through the command line override values, determine at what level to set them, and set them if possible

        *Args*:

            payload: current payload dictionary
            args: command line override args

        *Returns*:

            payload: updated payload dictionary with override values set

        """
        skip_keys = ['config', 'help']
        for k, v in vars(args).items():
            # If the name has a . then we are at the class level so we need to get the dict and check
            if len(k.split('.')) > 1:
                dict_key = k.split('.')[0]
                val_name = k.split('.')[1]
                if k not in skip_keys and v is not None:
                    # Handle bool types slightly differently as they are store_true
                    if isinstance(vars(args)[k], bool):
                        if vars(args)[k] is not False:
                            payload = self._dict_payload_override(payload, dict_key, val_name, v)
                    else:
                        payload = self._dict_payload_override(payload, dict_key, val_name, v)
            # else search the first level
            else:
                # Override the value in the payload if present
                if k not in skip_keys and v is not None:
                    # Handle bool types slightly differently as they are store_true
                    if isinstance(vars(args)[k], bool):
                        if vars(args)[k] is not False:
                            payload.update({k: v})
                    else:
                        payload.update({k: v})
        return payload

    @staticmethod
    def _dict_payload_override(payload, dict_key, val_name, value):
        """Updates the payload at the dictionary level

        First checks to see if there is an existing dictionary to insert into, if not creates an empty one. Then it
        inserts the updated value at the correct dictionary level

        Args:
            payload: current payload dictionary
            dict_key: dictionary key to check
            val_name: value name to update
            value: value to update

        Returns:

            payload: updated payload dictionary

        """
        if not hasattr(payload, dict_key):
            payload.update({dict_key: {}})
        payload[dict_key].update({val_name: value})
        return payload



# -*- coding: utf-8 -*-
"""Spock Setup"""

from pkg_resources import parse_requirements
import setuptools
import versioneer

with open('README.md', 'r') as fid:
    long_description = fid.read()

with open('REQUIREMENTS.txt', 'r') as fid:
    install_reqs = [str(req) for req in parse_requirements(fid)]

setuptools.setup(
    name='spock-config',
    description='Spock is a framework designed to help manage complex parameter configurations for Python applications',
    long_description=long_description,
    long_description_content_type='text/markdown',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="FMR LLC",
    url="https://github.com/fidelity/spock",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    project_urls={
        "Source": "https://github.com/fidelity/spock",
        "Documentation": "https://fidelity.github.io/spock/"
    },
    keywords="dataclass config argparser parameter configuration",
    packages=setuptools.find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    python_requires='>=3.6',
    install_requires=install_reqs
)



# -*- coding: utf-8 -*-

# Copyright 2019 FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles the definitions of arguments types for Spock (backend: dataclass)"""

from typing import List
from typing import Tuple
from typing import TypeVar


class BoolArg(int):
    """Spock boolean argument

    Overloads the bool type for Spock

    """
    def __new__(cls, x):
        """Creates a new instance of a Spock boolean arg

        *Args*:

            x: boolean value

        *Returns*:

            boolean variable
        """
        return super().__new__(cls, bool(x))


class IntArg(int):
    """Spock integer argument

    Overloads the integer type for Spock

    """
    def __new__(cls, x):
        """Creates a new instance of a Spock integer arg

        *Args*:

            x: integer value

        *Returns*:

            integer variable
        """
        return super().__new__(cls, x)


class IntOptArg(int):
    """Spock integer optional argument

    Overloads the integer type and makes the argument optional for Spock

    """
    def __new__(cls, x):
        """Creates a new instance of a Spock optional integer arg

        *Args*:

            x: integer value

        *Returns*:

            integer variable
        """
        return super().__new__(cls, x)


class FloatArg(float):
    """Spock float argument

    Overloads the float type for Spock

    """
    def __new__(cls, x):
        """Creates a new instance of a Spock float arg

        *Args*:

            x: float value

        *Returns*:

            float variable
        """
        return super().__new__(cls, x)


class FloatOptArg(float):
    """Spock float optional argument

    Overloads the float type and makes the argument optional for Spock

    """
    def __new__(cls, x):
        """Creates a new instance of a Spock float optional arg

        *Args*:

            x: float value

        *Returns*:

            float variable
        """
        return super().__new__(cls, x)


class StrArg(str):
    """Spock string argument

    Overloads the string type  for Spock

    """
    def __new__(cls, x):
        """Creates a new instance of a Spock string arg

        *Args*:

            x: string value

        *Returns*:

            string variable
        """
        return super().__new__(cls, x)


class StrOptArg(str):
    """Spock string optional argument

    Overloads the string type and makes the argument optional for Spock

    """
    def __new__(cls, x):
        """Creates a new instance of a Spock string optional arg

        *Args*:

            x: string value

        *Returns*:

            string variable
        """
        return super().__new__(cls, x)


# Make a type var
__T = TypeVar('__T')


class ListArg(List[__T]):  # pylint: disable=too-few-public-methods
    """Spock list argument

    Overloads the list type for Spock

    """
    @staticmethod
    def defaults(values: List):
        """Creates a new instance of a Spock list arg

        *Args*:

            values: list values

        *Returns*:

            list variable
        """
        return values


class ListOptArg(List[__T]):  # pylint: disable=too-few-public-methods
    """Spock list optional argument

    Overloads the list type and makes the argument optional for Spock

    """
    @staticmethod
    def defaults(values: List):
        """Creates a new instance of a Spock list optional arg

        *Args*:

            values: list values

        *Returns*:

            list variable
        """
        return values


class TupleArg(Tuple[__T]):  # pylint: disable=too-few-public-methods
    """Spock tuple argument

    Overloads the tuple type for Spock

    """
    @staticmethod
    def defaults(values: Tuple):
        """Creates a new instance of a Spock tuple arg

        *Args*:

            values: tuple values

        *Returns*:

            tuple variable
        """
        return values


class TupleOptArg(Tuple[__T]):  # pylint: disable=too-few-public-methods
    """Spock tuple optional argument

    Overloads the tuple type and makes the argument optional for Spock

    """
    @staticmethod
    def defaults(values: Tuple):
        """Creates a new instance of a Spock tuple optional arg

        *Args*:

            values: tuple values

        *Returns*:

            tuple variable
        """
        return values


class ChoiceArg:  # pylint: disable=too-few-public-methods
    """Spock type of a choice set

    Requires default or value to be from a predefined set

    """
    def __init__(self, choice_set: List, default=None):
        self.choice_set = choice_set
        self.default = default
        self.set_type = self._verify()

    def _verify(self):
        """Validates the types within a set

        *Returns*:

            str name of type

        """
        type_set = {type(val) for val in self.choice_set}
        if len(type_set) > 1:
            raise TypeError(f'ChoiceArg must all be of the same type: {type_set}')
        return list(type_set)[0]


class SavePathOptArg(str):
    """Spock special key for saving the Spock config to file

    Defines a special key use to save the current Spock config to file

    """
    def __new__(cls, x):
        return super().__new__(cls, x)


def boolean_string(bool_string):
    """Map boolean string to boolean type

    *Args*:

        s: boolean string

    *Returns*:

        string True/False to actual bool type

    """
    if bool_string not in {'False', 'True'}:
        raise ValueError('Not a valid boolean string')
    return bool_string == 'True'



# -*- coding: utf-8 -*-

# Copyright 2019 FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Creates the spock config decorator that wraps dataclasses - now an adapter to attr"""

import enum
from inspect import isfunction
from typing import List
from typing import Optional
from typing import Tuple
import attr
from spock.backend.attr.typed import katra
from spock.backend.attr.typed import SavePath
from spock.backend.dataclass._dataclasses import dataclass


def _enum_adapter(key, obj):
    """Adapter for ChoiceSet to Enum

    *Args*:

        key: name of enum
        obj: values in the ChoiceSet

    Returns:

        enum_obj: constructed enum object

    """
    enum_set = {('option_' + str(idx)): val for idx, val in enumerate(obj.choice_set)}
    enum_obj = enum.Enum(key, enum_set)
    return enum_obj


def _list_adapter(_, obj):
    """Adapter for List types

    *Args*:

        _: unused
        obj: old list type

    Returns:

        List type

    """
    return List[obj.__args__[0]]


def _list_optional_adapter(_, obj):
    """Adapter for Optional List types

    *Args*:

        _: unused
        obj: old list type

    Returns:

        Optional List type

    """
    return Optional[List[obj.__args__[0]]]


def _tuple_adapter(_, obj):
    """Adapter for Tuple types

    *Args*:

        _: unused
        obj: old tuple type

    Returns:

        Tuple type

    """
    return Tuple[obj.__args__[0]]


def _tuple_optional_adapter(_, obj):
    """Adapter for Optional Tuple types

    *Args*:

        _: unused
        obj: old tuple type

    Returns:

        Optional Tuple type

    """
    return Optional[Tuple[obj.__args__[0]]]


TYPE_MAP = {
    'BoolArg': bool,
    'IntArg': int,
    'IntOptArg': Optional[int],
    'FloatArg': float,
    'FloatOptArg': Optional[float],
    'StrArg': str,
    'StrOptArg': Optional[str],
    'ListArg': _list_adapter,
    'ListOptArg': _list_optional_adapter,
    'TupleArg': _tuple_adapter,
    'TupleOptArg': _tuple_optional_adapter,
    'SavePathOptArg': SavePath,
    'ChoiceArg': _enum_adapter
}


def spock_legacy_dataclass(*args, **kwargs):
    """Wrapper to dataclass that forms the base of Spock configs
    *Args*:
        *args:
        **kwargs:
    *Returns*:
        frozen dataclass: frozen version of the dataclass
    """
    kwargs['frozen'] = True
    return dataclass(*args, **kwargs)


def spock_dataclass(*args, **kwargs):
    """Wrapper to dataclass that forms the base of Spock configs

    *Args*:

        *args:
        **kwargs:

    *Returns*:

        frozen dataclass: frozen version of the dataclass
    """
    cls = args[0]
    # Use the adaptor to convert into the attr class
    attrs_dict, bases = _adapter(cls=cls)
    return attr.make_class(name=cls.__name__, bases=bases, attrs=attrs_dict, kw_only=True, frozen=True)


def _adapter(cls):
    """Takes a class an adapts the dataclass backend to the attr backend

    Maps the old interface and backend of dataclasses to the new interface and backend of attrs. Based on a type map
    dictionary and mapping functions it provides the ability to 1:1 map between inferfaces.

    *Args*:

        cls: input class

    *Returns*:

        attrs_dict: a dictionary of current attributes to make
        bases: any base classes to inherit from

    """
    # Make a blank attrs dict for new attrs
    attrs_dict = {}
    # We are mapping to the attr backend thus we need to get the parent classes for inheritance
    # We do this by using the mro and grabbing anything that is not the first and last indices in the list and wrapping
    # it into a tuple
    if len(cls.__mro__[1:-1]) > 0:
        bases = tuple(cls.__mro__[1:-1])
    # if there are not parents pass a blank tuple
    else:
        bases = ()
    if hasattr(cls, '__annotations__'):
        for k, v in cls.__annotations__.items():
            # If the cls has the attribute then a default was set
            if hasattr(cls, k):
                default = getattr(cls, k)
            elif hasattr(v, 'default'):
                default = getattr(v, 'default')
            else:
                default = None
            if hasattr(v, '__name__'):
                typed = TYPE_MAP.get(v.__name__)
            elif hasattr(v, '__origin__'):
                typed = TYPE_MAP.get(v.__origin__.__name__)
            else:
                typed = TYPE_MAP.get(type(v).__name__)
            if isfunction(typed):
                typed = typed(k, v)
            attrs_dict.update({k: katra(typed=typed, default=default)})
    return attrs_dict, bases



# -*- coding: utf-8 -*-

# Copyright 2019 FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles the building/saving of the configurations from the Spock config classes"""

import sys
from typing import Generic
from spock.backend.dataclass._dataclasses import is_dataclass
from spock.backend.base import BaseBuilder
from spock.backend.dataclass.utils import cast
minor = sys.version_info.minor
if minor < 7:
    from typing import GenericMeta as _GenericAlias
else:
    from typing import _GenericAlias


class DataClassBuilder(BaseBuilder):
    """Dataclass specific builder

    Class that handles building for the dataclass backend

    *Attributes*

        input_classes: list of input classes that link to a backend
        _configs: None or List of configs to read from
        _create_save_path: boolean to make the path to save to
        _desc: description for the arg parser
        _no_cmd_line: flag to force no command line reads
        save_path: list of path(s) to save the configs to

    """
    def __init__(self, *args, configs=None, create_save_path=False, desc='', no_cmd_line=False, **kwargs):
        super().__init__(*args, configs=configs, create_save_path=create_save_path, desc=desc,
                         no_cmd_line=no_cmd_line, **kwargs)
        self._optional_types = {'FloatOptArg', 'IntOptArg', 'StrOptArg',
                                'ListOptArg', 'TupleOptArg', 'SavePathOptArg'}
        for arg in self.input_classes:
            if not is_dataclass(arg):
                raise TypeError('*arg inputs to ConfigArgBuilder must all be instances of dataclass')

    def print_usage_and_exit(self, msg=None, sys_exit=True):
        print('USAGE:')
        print(f'  {sys.argv[0]} -c [--config] config1 [config2, config3, ...]')
        print('CONFIG:')
        for data_class in self.input_classes:
            print('  ' + data_class.__name__ + ':')
            dc_vars = vars(data_class)
            for key, val in dc_vars.get('__dataclass_fields__').items():
                if type(val.type).__name__ == 'ChoiceArg':
                    type_name = type(val.type).__name__
                # Magic again -- check for type == type allows for delineation between basic types and list/tuples
                elif type(val.type) == type:
                    type_name = val.type.__name__
                else:
                    type_name = val.type.__origin__.__name__
                    type_name += '[' + self._extract_base_type(val.type).__name__ + ']'
                print(f'    {key}: {type_name}')
        if msg is not None:
            print(msg)
        if sys_exit:
            sys.exit(1)

    def _handle_arguments(self, args, class_obj):
        fields = {}
        # Access the vars
        dc_vars = vars(class_obj)
        # Get the dataclass name
        dc_name = class_obj.__name__
        for key, val in dc_vars.get('__dataclass_fields__').items():
            # pure magic -- Lists, Tuples, etc. are not of type type (they are GenericAlias) so one must
            # check against this before accessing the __name__ attribute which GenericAlias does not have
            if type(val.type) == type:
                if val.type.__name__ == 'SavePathOptArg':
                    self.save_path = args.get(key)
            # Check if namespace is named and then check for key -- checking for local def
            if dc_name in args and key in args[dc_name]:
                fields[key] = self._check_function(args[dc_name][key], val)
            # If not named then just check for keys -- checking for global def
            elif key in args:
                fields[key] = self._check_function(args[key], val)
            # If not found then fallback on defaults if defined
            else:
                default, found_default = self.check_for_defaults(val)
                if not found_default:
                    if type(val.type) == type and val.type.__name__ in self._optional_types:
                        fields[key] = None
                        continue
                    elif type(val.type) != type and val.type.__origin__.__name__ in self._optional_types:
                        fields[key] = None
                        continue
                    elif 'Bool' in val.type.__name__:
                        fields[key] = False
                        continue
                    else:
                        raise ValueError(f'Required value {dc_name}.{key}: no default set or value defined in file')
                fields[key] = self._check_function(default, val)
        return fields

    @staticmethod
    def _int_to_float(inst, target_type):
        """Converts instance int to float

        *Args*:
            inst: instance
            target_type: target type

        *Returns*:
            inst: instance type cast into float

        """
        if target_type == float and type(inst) == int:
            inst = float(inst)
        return inst

    def _check_function(self, x, val):
        """Wrapper around the valid type check with a cast

        *Args*:

            x: instance
            val: value

        *Returns*:

            casted value

        """
        return cast(self._check_valid_type(x, val))

    def _check_valid_type(self, instance, val):
        """Checks that the instance is of the correct type

        *Args*:

            instance: object instance
            val: value

        *Returns*:

            instance: object instance

        """
        if type(val.type) == type:
            # pure magic -- Lists, Tuples, etc. are not of type type (they are GenericAlias) so one must
            # check against this before accessing the __name__ attribute which GenericAlias does not have
            # Get the base variable type
            var_type = val.type.__bases__[0]
            instance = self._int_to_float(instance, var_type)
            valid = isinstance(instance, var_type)
            if not valid:
                raise ValueError(f'Wrong type ({type(instance)}) passed to {val.name}. Require {var_type}')
        elif type(val.type).__name__ == 'ChoiceArg':
            instance = self._check_choice_type(val.type, instance)
            var_type = val.type.set_type
            valid = isinstance(instance, var_type)
            if not valid:
                raise ValueError(f'Wrong type ({type(instance)}) passed to {val.name}. Require {var_type}')
        else:
            # It's an iterable - check it's a list (only iterable provided by markup)
            var_type = self._extract_base_type(val.type)
            iter_name = val.type.__origin__.__name__
            valid = isinstance(instance, (tuple, list))
            if not valid:
                raise ValueError(
                    f'Wrong type ({type(instance).__name__}) passed to {val.name}.\n'
                    f'Require {iter_name}[{var_type.__name__}]')
            instance = tuple((self._int_to_float(i, var_type) for i in instance))
            if len(instance) > 0:
                valid = isinstance(instance[0], var_type)
                if not valid:
                    raise ValueError(
                        f'Wrong type (List[{type(instance[0]).__name__}]) passed to {val.name}.\n'
                        f'Require {iter_name}[{var_type.__name__}]')
        return instance

    @staticmethod
    def _extract_base_type(given_type):
        """Extracts the type from a _GenericAlias

        *Args*:

            tp: type

        *Returns*:

            tp: type of generic type
        """
        if isinstance(given_type, _GenericAlias) and given_type is not Generic:
            return given_type.__args__[0]  # assume we only have generic types with a single argument
        return given_type

    def check_for_defaults(self, val):
        """Checks for default values

        *Args*:

            default: default value
            default_factory: default factory

        *Returns*:

            default_val: value of default
            found_default: boolean if default found

        """
        found_default = False
        default_val = None
        if type(val.type).__name__ == 'ChoiceArg':
            default_val = self._check_choice_type(val.type, val.type.default)
            found_default = True
        elif type(val.default).__name__ != '_MISSING_TYPE':
            default_val = val.default
            found_default = True
        elif type(val.default_factory).__name__ != '_MISSING_TYPE':
            default_val = val.default_factory()
            found_default = True
        return default_val, found_default

    @staticmethod
    def _check_choice_type(choice_set, val):
        """Checks the type and set of a ChoiceArg value

        *Args*:

            choice_set: ChoiceArg instance
            val: value to set
            val_name: name of the parameter

        *Returns*:

            val: value to set

        """
        if val not in choice_set.choice_set:
            raise ValueError(f'{val} is not within the set of defined choices {choice_set.choice_set}')
        return val
